library(ggplot2)
library(reshape2)
library(boot)
library(matrixStats)
library(dplyr)
library(tidyr)
library(Hmisc)
library(grid)
library(xtable)
library(lme4)
library(lmerTest)

theme <- theme_bw() + 
  theme(plot.title = element_text(face = "bold", hjust = 0.5),
        axis.title.x = element_text(color = "forestgreen", face = "bold"),
        axis.title.y = element_text(color = "forestgreen", face = "bold"),
        legend.title = element_text(color = "forestgreen", face = "bold"),
        panel.grid.minor.x = element_blank(),
        panel.grid.major.x = element_blank(), aspect.ratio = 1)

theme_set(theme)

###################################################################################################################
# Functions
###################################################################################################################
multiplot <- function(..., plotlist=NULL, file, cols=1, layout=NULL) 
{
  library(grid)
  
  # Make a list from the ... arguments and plotlist
  plots <- c(list(...), plotlist)
  
  numPlots = length(plots)
  
  # If layout is NULL, then use 'cols' to determine layout
  if (is.null(layout)) 
  {
    # Make the panel
    # ncol: Number of columns of plots
    # nrow: Number of rows needed, calculated from # of cols
    layout <- matrix(seq(1, cols * ceiling(numPlots/cols)),
                     ncol = cols, nrow = ceiling(numPlots/cols))
  }
  
  if (numPlots==1)
  {
    print(plots[[1]])
    
  } 
  else 
  {
    # Set up the page
    grid.newpage()
    pushViewport(viewport(layout = grid.layout(nrow(layout), ncol(layout))))
    
    # Make each plot, in the correct location
    for (i in 1:numPlots)
    {
      # Get the i,j matrix positions of the regions that contain this subplot
      matchidx <- as.data.frame(which(layout == i, arr.ind = TRUE))
      
      print(plots[[i]], vp = viewport(layout.pos.row = matchidx$row,
                                      layout.pos.col = matchidx$col))
    }
  }
}

inverseLogit <- function(x)
{
  log(x) - log(1 - x)
}

substitution_distance <- function(s1, s2) 
{ 
  mapply(function(c1, c2) sum(c1 != c2), strsplit(s1, ''), strsplit(s2, ''))
}

###################################################################################################################
# Data re-shaping
###################################################################################################################
# Working directory should be set to the same directory as analysis.R
#setwd("Certainty Code and Data\Experiment 2")

bestFitData <- read.csv('bestFitsStudy2.csv')

modelData <- read.csv('modelDataStudy2.csv')
#modelData <- read.csv('modelDataStudy2Simplified.csv')

data <- read.csv("behavioralDataStudy2.csv")

data$Generalization <- sapply(data$Generalization, function(x) sub("'", '', x))
data$Correct <- sapply(data$Correct, function(x) sub("'", '', x))

data$totalCorrect <- abs(substitution_distance(data$Generalization, data$Correct) - 8)
data$percentAccuracy <- abs((substitution_distance(data$Generalization, data$Correct) - 8) / 8)

data$firstAccuracy <- abs(substitution_distance(substr(data$Generalization, 1, 1), substr(data$Correct, 1, 1)) - 1)
data$secondAccuracy <- abs(substitution_distance(substr(data$Generalization, 2, 2), substr(data$Correct, 2, 2)) - 1)
data$thirdAccuracy <- abs(substitution_distance(substr(data$Generalization, 3, 3), substr(data$Correct, 3, 3)) - 1)
data$fourthAccuracy <- abs(substitution_distance(substr(data$Generalization, 4, 4), substr(data$Correct, 4, 4)) - 1)
data$fifthAccuracy <- abs(substitution_distance(substr(data$Generalization, 5, 5), substr(data$Correct, 5, 5)) - 1)
data$sixthAccuracy <- abs(substitution_distance(substr(data$Generalization, 6, 6), substr(data$Correct, 6, 6)) - 1)
data$seventhAccuracy <- abs(substitution_distance(substr(data$Generalization, 7, 7), substr(data$Correct, 7, 7)) - 1)
data$eighthAccuracy <- abs(substitution_distance(substr(data$Generalization, 8, 8), substr(data$Correct, 8, 8)) - 1)

data <- gather(data, "Trial", "trialCorrect", 14:21, factor_key = TRUE)
data$Trial <- recode(data$Trial, "firstAccuracy" = "1", "secondAccuracy" = "2", "thirdAccuracy" = "3",
                                 "fourthAccuracy" = "4", "fifthAccuracy" = "5", "sixthAccuracy" = "6",
                                 "seventhAccuracy" = "7", "eighthAccuracy" = "8")

data <- data %>% group_by(Trial, Condition) %>% mutate(BehavioralAccuracy = mean(trialCorrect), BehavioralCertainty = mean(Certainty))

# Creation of count file
#counts <- ddply(data, .(Condition, Trial), function(df){data.frame(Correct = sum(df$trialCorrect), Incorrect = sum(df$trialCorrect == 0), Total = sum(df$trialCorrect) + sum(df$trialCorrect == 0))})
#write.csv(counts, "Study2Counts.csv", row.names = FALSE)

data <- merge(modelData, data)

data$PredictedAccuracyBins <- cut(data$PredictedAccuracy, breaks = c(0, .1, .2, .3, .4, .5, .6, .7, .8, .9, 1.1), labels = 0:9)
data$HypothesisAccuracyBins <- cut(data$HypothesisAccuracy, breaks = c(0, .1, .2, .3, .4, .5, .6, .7, .8, .9, 1.1), labels = 0:9)
data$BehavioralCertaintyBins <- cut(data$BehavioralCertainty, breaks = c(0, .1, .2, .3, .4, .5, .6, .7, .8, .9, 1.1), labels = 0:9)

meanData <- data %>% group_by(Trial, Condition) %>% summarise(totalCorrect = mean(totalCorrect),
                                                              Certainty = mean(Certainty),
                                                              Entropy = mean(Entropy),
                                                              MAP = mean(MAP),
                                                              MAPNoPrior = mean(MAPNoPrior),
                                                              DomainEntropy = mean(DomainEntropy),
                                                              PredictedAccuracy = mean(PredictedAccuracy),
                                                              trialCorrect = mean(trialCorrect))

###################################################################################################################
# Visualizations
###################################################################################################################
# Behavioral Accuracy Binned by Model Accuracy
ggplot(data, aes(x = as.numeric(as.character(cut2(PredictedAccuracy, g = 5, levels.mean = TRUE))), y = BehavioralAccuracy)) + 
  stat_summary(fun.y = mean, geom = "point", shape = 5, size = 5, color = "blue") +
  stat_summary(fun.data = mean_cl_boot, geom = "pointrange", color = "blue") +
  geom_point(aes(x = PredictedAccuracy, y = BehavioralAccuracy), alpha = .005) +
  expand_limits(x = 0, y = 0) +
  expand_limits(x = 1, y = 1) +
  scale_x_continuous(expand = c(0, 0)) + 
  scale_y_continuous(expand = c(0, 0)) +
  geom_abline(slope = 1, linetype = "dotted") +
  theme(axis.title = element_text(size = 14, face = "bold")) +
  labs(x = "Model Accuracy", y = "Behavioral Accuracy") +
  ggtitle("Model Predicts Accuracy")

# Condition vs. Certainty
ggplot(data, aes(Condition, Certainty)) +
  geom_jitter(position = position_jitter(w = .1, h = .1)) +
  ggtitle("Condition vs. Certainty")

# Condition vs. Certainty RT
ggplot(data, aes(Condition, CertaintyRT)) +
  geom_jitter(position = position_jitter(w = .1, h = .1)) +
  geom_boxplot() +
  #ylim(30000, 50000) +
  ggtitle("Condition vs. Certainty RT")

ggplot(data, aes(Certainty, GeneralizationRT)) +
  geom_boxplot() +
  geom_smooth(method = "lm", alpha = 0.1) +
  geom_jitter(position = position_jitter(w = .2, h = .2)) +
  ggtitle("Generalization RT vs. certainty")

# Condition vs. Total Correct
ggplot(data, aes(Condition, totalCorrect), fill = certainty) +
  geom_boxplot() +
  stat_summary(fun.y = mean, geom = 'point', shape = 1, size = 4, color = "blue") +
  geom_jitter(position = position_jitter(w = .2, h = .2)) +
  ggtitle("Condition vs. Total Correct")

# Condition vs. Generalization RT
ggplot(data, aes(Condition, GeneralizationRT)) +
  geom_jitter(position = position_jitter(w = .1, h = .1)) +
  geom_boxplot() +
  ggtitle("Condition vs. Generalization RT")

# Accuracy vs. Certainty
ggplot(data, aes(totalCorrect, Certainty)) +
  geom_smooth(method = "lm", alpha = 0.1) +
  geom_jitter(position = position_jitter(w = .2, h = .2)) +
  ggtitle("Accuracy vs. Certainty")

ggplot(data, aes(totalCorrect, GeneralizationRT)) +
  geom_smooth(method = "lm", alpha = 0.1) +
  geom_jitter(position = position_jitter(w = .2, h = .2)) +
  ggtitle("Generalization RT vs. Total Correct")

ggplot(data, aes(PredictedAccuracy, trialCorrect)) +
  geom_smooth(method = "lm", alpha = 0.1) +
  geom_jitter(position = position_jitter(w = .2, h = .2)) +
  ggtitle("Predicted Accuracy vs. Accuracy")

# Certainty vs. Total Correct
v3 <- ggplot(meanData, aes(x = as.numeric(as.character(cut2(totalCorrect, g = 5, levels.mean = TRUE))), y = Certainty)) + 
  stat_summary(fun.y = mean, geom = "point", shape = 5, size = 5, color = "blue") +
  stat_summary(fun.data = mean_cl_boot, geom = "pointrange", color = "blue") +
  geom_point(aes(x = totalCorrect, y = Certainty), alpha = .1) +
  #ggtitle(expression(R^{2} ~ " = .44")) +
  ggtitle(expression("Experiment 2 - " ~ R^{2} ~ " = .44")) +
  labs(x = "Total Correct", y = "Certainty") +
  expand_limits(y = 0, y = 1) +
  scale_y_continuous(expand = c(0, 0))

# Certainty vs. Domain Entropy
v4 <- ggplot(meanData, aes(x = as.numeric(as.character(cut2(DomainEntropy, g = 5, levels.mean = TRUE))), y = Certainty)) +
  stat_summary(fun.y = mean, geom = "point", shape = 5, size = 5, color = "blue") +
  stat_summary(fun.data = mean_cl_boot, geom = "pointrange", color = "blue") +
  geom_point(aes(x = DomainEntropy, y = Certainty), alpha = .1) +
  #ggtitle(expression(R^{2} ~ " = .49")) +
  ggtitle(expression("Experiment 2 - " ~ R^{2} ~ " = .49")) +
  labs(x = "Domain Entropy", y = "Certainty") +
  expand_limits(y = 0, y = 1) +
  scale_y_continuous(expand = c(0, 0))

# Certainty vs. Entropy (Ideal Learner) ***A***
v3 <- ggplot(meanData, aes(x = as.numeric(as.character(cut2(Entropy, g = 5, levels.mean = TRUE))), y = Certainty)) +
  stat_summary(fun.y = mean, geom = "point", shape = 5, size = 5, color = "blue") +
  stat_summary(fun.data = mean_cl_boot, geom = "pointrange", color = "blue") +
  geom_point(aes(x = Entropy, y = Certainty), alpha = .1) +
  ggtitle(expression(R^{2} ~ " = .24")) +
  labs(x = "Entropy", y = "Certainty") +
  expand_limits(y = 0, y = 1) +
  scale_y_continuous(expand = c(0, 0))

# Certainty By MAP (learners who only have a single hypothesis in mind) ***D***
v4 <- ggplot(meanData, aes(x = as.numeric(as.character(cut2(MAP, g = 5, levels.mean = TRUE))), y = Certainty)) + 
  stat_summary(fun.y = mean, geom = "point", shape = 5, size = 5, color = "blue") +
  stat_summary(fun.data = mean_cl_boot, geom = "pointrange", color = "blue") +
  geom_point(aes(x = MAP, y = Certainty), alpha = .1) +
  ggtitle(expression(R^{2} ~ " = .19")) +
  labs(x = "MAP Hypothesis", y = "Certainty") +
  expand_limits(y = 0, y = 1) +
  scale_y_continuous(expand = c(0, 0))

# Certainty By Log Maximum Likelihood (learners without a simplicity bias) ***E***
v5 <- ggplot(meanData, aes(x = as.numeric(as.character(cut2(log(MAPNoPrior), g = 5, levels.mean = TRUE))), y = Certainty)) + 
     stat_summary(fun.y = mean, geom = "point", shape = 5, size = 5, color = "blue") +
     stat_summary(fun.data = mean_cl_boot, geom = "pointrange", color = "blue") +
     geom_point(aes(x = log(MAPNoPrior), y = Certainty), alpha = .1) +
     ggtitle(expression(R^{2} ~ " = .32")) +
     labs(x = "Log Max Likelihood", y = "Certainty") +
     expand_limits(y = 0, y = 1) +
     scale_y_continuous(expand = c(0, 0))

multiplot(v1, v3, v5, v2, v4, cols = 2)

# Certainty and Accuracy by Condition
ggplot(data, aes(x = Condition, y = Certainty)) + 
  stat_summary(fun.y = mean, geom = "point", shape = 1, color = "blue", size = 4) + 
  stat_summary(fun.data = mean_cl_boot, geom = "linerange") +
  stat_summary(aes(x = Condition, y = percentAccuracy), fun.y = mean, geom = "point", color = "red", size = 4) + 
  stat_summary(aes(x = Condition, y = percentAccuracy), fun.data = mean_cl_boot, geom = "linerange") +
  ggtitle("Certainty and Accuracy by Condition") +
  labs(color = "Condition", y = "") +
  ylim(0, 1)

# Certainty and Accuracy by Subject
ggplot(data, aes(x = Trial, y = Certainty)) + 
  stat_summary(fun.y = mean, geom = "point", shape = 5, size = 5, fill = "blue") + 
  stat_summary(fun.data = mean_cl_boot, geom = "pointrange", color = "blue") +
  stat_summary(aes(x = Trial, y = trialCorrect), fun.y = mean, geom = "point", shape = 7, size = 5, color = "red") + 
  stat_summary(aes(x = Trial, y = trialCorrect), fun.data = mean_cl_boot, geom = "pointrange", color = "red") +
  ggtitle("Certainty and Accuracy by Condition") +
  labs(color = "Condition", y = "") +
  scale_color_manual(values = c("Women" = '#ff00ff','Men' = '#3399ff')) +
  ylim(0, 1) +
  facet_wrap(~ID)

# Behavioral Accuracy By Model Accuracy
ggplot(modelData, aes(x = PredictedAccuracy, y = BehavioralAccuracy)) + 
  stat_summary(fun.y = mean, geom = "point", shape = 1, size = 4, color = "blue") +
  ggtitle("Behavioral Accuracy By Model Accuracy") +
  geom_abline(slope = 1, intercept = 0) +
  xlim(0, 1) +
  facet_wrap(~Condition)

# Behavioral Accuracy Binned by Model Accuracy
ggplot(modelData, aes(x = HypothesisAccuracyBins, y = BehavioralAccuracy)) + 
  geom_boxplot() + 
  stat_summary(fun.y = mean, geom = "point", shape = 5, size = 4) + 
  stat_summary(fun.data = mean_cl_boot, geom = "pointrange") +
  ylim(0, 1) +
  ggtitle("Behavioral Accuracy Binned by Model Accuracy") +
  geom_point() + 
  geom_jitter(position = position_jitter(w = .4, h = .1)) +
  facet_wrap(~Condition)

# Accuracy by Condition/Trial
ggplot(data, aes(Trial, ModelAccuracy)) +
  stat_summary(fun.y = mean, geom='point', shape = 1, size = 4, color = "blue") +
  stat_summary(aes(Trial, BehavioralAccuracy), fun.y = mean, geom='point', shape = 2, size = 4, color = "red") +
  ggtitle("Model (blue) and Behavioral (red) Accuracy by Condition") +
  labs(x = "Trial", y = "Accuracy") +
  scale_x_continuous(breaks = seq(0, 24, by = 1)) +
  facet_wrap(~Condition)

# Certainty by Trial
ggplot(modelData, aes(Trial, BehavioralCertainty)) +
  stat_summary(fun.y = mean, geom='point', shape = 2, size = 4, color = "red") +
  ggtitle("Behavioral Certainty by Trial") +
  labs(x = "Trial", y = "Certainty") +
  scale_x_continuous(breaks = seq(0, 24, by = 1)) +
  ylim(0, 1) +
  facet_wrap(~Condition)

# Proportions of Accuracy and Certainty
ggplot(data, aes(BehavioralCertainty, BehavioralAccuracy)) +
  stat_summary(fun.y = mean, geom='point') +
  ggtitle("Proportions of Accuracy and Certainty") +
  labs(x = "Certainty", y = "Accuracy") +
  xlim(0, 1) +
  ylim(0, 1) +
  geom_abline(slope = 1, intercept = 0)
  #facet_wrap(~Condition)

# Behavioral Accuracy Binned by Model Accuracy
ggplot(data, aes(x = ModelAccuracyBins, y = BehavioralAccuracy)) + 
  geom_boxplot() + 
  geom_point() + 
  geom_jitter(position = position_jitter(w = .4, h = .1)) +
  stat_summary(fun.y = mean, geom = "point", shape = 5, size = 4) + 
  stat_summary(fun.data = mean_cl_boot, geom = "pointrange") +
  ylim(0, 1) +
  ggtitle("Behavioral Accuracy Binned by Model Accuracy")

# Behavioral Certainty Binned by Model Accuracy
ggplot(data, aes(x = ModelAccuracyBins, y = BehavioralCertainty)) + 
  geom_boxplot() + 
  geom_point() + 
  geom_jitter(position = position_jitter(w = .4, h = .1)) +
  stat_summary(fun.y = mean, geom = "point", shape = 5, size = 4) + 
  stat_summary(fun.data = mean_cl_boot, geom = "pointrange") +
  ylim(0, 1) +
  ggtitle("Behavioral Certainty Binned by Model Accuracy")

# Certainty by Consecutive Correct
ggplot(data, aes(x = ConsecutiveCorrect, y = certainty)) + 
  geom_point() + 
  geom_jitter(position = position_jitter(w = .4, h = .1)) +
  stat_summary(fun.y = mean, geom = "point", shape = 5, size = 4, fill = "red") + 
  stat_summary(fun.data = mean_cl_boot, geom = "pointrange", color = "red") +
  ggtitle("Certainty by Consecutive Correct") +
  labs(color = "Condition", x = "Consecutive Correct", y = "Certainty") +
  ylim(0, 1)

# Certainty and Accuracy by Trial
ggplot(modelData, aes(x = Trial, y = BehavioralCertainty)) + 
  stat_summary(fun.y = mean, geom = "point", shape = 5, size = 4, fill = "blue") + 
  stat_summary(fun.data = mean_cl_boot, geom = "pointrange", color = "green") +
  stat_summary(aes(x = Trial, y = accuracy), fun.y = mean, geom = "point", shape = 1, size = 4, color = "red") + 
  stat_summary(aes(x = Trial, y = accuracy), fun.data = mean_cl_boot, geom = "errorbar", color = "red") +
  ggtitle("Certainty and Accuracy by Trial") +
  ylim(0, 1)

# Certainty and Accuracy when Concept == Red AND Small
ggplot(filter(data, Condition == 1), aes(x = Trial, y = certainty)) + 
  stat_summary(fun.y = mean, geom = "point", shape = 5, size = 4, fill = "blue") + 
  stat_summary(fun.data = mean_cl_boot, geom = "pointrange", color = "green") +
  stat_summary(aes(x = Trial, y = accuracy), fun.y = mean, geom = "point", shape = 1, size = 4, color = "red") + 
  stat_summary(aes(x = Trial, y = accuracy), fun.data = mean_cl_boot, geom = "errorbar", color = "red") +
  ggtitle("Certainty and Accuracy for AND") +
  ylim(0, 1)

# Certainty by Condition
ggplot(data, aes(x = Condition, y = certainty)) + 
  geom_boxplot() + 
  geom_point() + 
  geom_jitter(position = position_jitter(w = .4, h = .1)) +
  stat_summary(fun.y = mean, geom = "point", shape = 5, size = 4) + 
  stat_summary(fun.data = mean_cl_boot, geom = "pointrange") +
  ggtitle("Certainty by Condition")

# Accuracy by Condition
ggplot(data, aes(x = Condition, y = accuracy)) + 
  geom_boxplot() + 
  geom_point() + 
  geom_jitter(position = position_jitter(w = .4, h = .1)) +
  stat_summary(fun.y = mean, geom = "point", shape = 5, size = 4) + 
  stat_summary(fun.data = mean_cl_boot, geom = "pointrange") +
  ggtitle("Accuracy by Condition")

# RT by Trial
ggplot(data, aes(x = Trial, y = RT, color = Certainty)) +
  stat_summary(fun.y = mean, geom = "point") +
  geom_smooth(method = "lm", alpha = 0.1) + #, formula = y ~ x + Condition) + 
  ggtitle("RT by Trial") +
  facet_wrap(~Condition)

# RT by Condition
ggplot(data, aes(x = Condition, y = RT, fill = Certainty)) + 
  stat_summary(fun.y = mean, geom = "bar", position = position_dodge(.9)) +
  stat_summary(fun.data = mean_cl_boot, geom = "pointrange", position = position_dodge(.9)) +
  ggtitle("RT by Condition")

# RT/Certainty
ggplot(data, aes(x = Certainty, y = RT)) + 
  stat_summary(fun.y = mean, geom = "bar") +
  stat_summary(fun.data = mean_cl_boot, geom = "pointrange") +
  ggtitle("RT by Certainty")

# RT/Accuracy
ggplot(data, aes(x = Accuracy, y = RT, fill = Certainty)) + 
  stat_summary(fun.y = mean, geom = "bar", position = position_dodge(.9)) +
  stat_summary(fun.data = mean_cl_boot, geom = "pointrange", position = position_dodge(.9)) +
  ggtitle("RT by Accuracy")

###################################################################################################################
# Regressions
###################################################################################################################
accuracyNull <- glmer(trialCorrect ~ (1 | Condition), data, family = "binomial")
certaintyNull <- glmer(Certainty ~ (1 | Condition), data, family = "binomial")

modelPredictsAccuracy <- glmer(trialCorrect ~ scale(PredictedAccuracy) + (1 | Condition), data, family = "binomial")
summary(modelPredictsAccuracy)
cor(meanData$trialCorrect, meanData$PredictedAccuracy) ^ 2

summary(glm(Certainty ~ trialCorrect, data, family = binomial()))

summary(glm(Certainty ~ Gender, data, family = binomial()))


summary(glm(RT ~ Certainty, data))

summary(lmer(RT ~ Certainty + (1 | ID), data))

summary(lmer(RT ~ accuracy + (1 | ID), data))

summary(lmer(RT ~ NumberAccurate + (1 | ID), data))

# Try with CERTAINTY and accuracy
summary(lmer(RT ~ certainty * accuracy * Trial + (1 | ID), data))

summary(glmer(certainty ~ accuracy * Trial + (1 | ID), data, family = "binomial"))

summary(glmer(accuracy ~ certainty * Trial + (1 | ID), data, family = "binomial"))

#Model predicts behaviorial accuracy?
summary(glmer(trialCorrect ~ PredictedAccuracy + (1 | ID), data, family = "binomial"))

# Subject reliability
#i <- 1
#experimentReliability <- 0

#while(i <= 1000) 
#{
#  data1 <- data %>% sample_n(nrow(data) / 2, replace = FALSE)
#  data2 <- anti_join(data, data1)
#  temp <- data1 %>% group_by(Trial, Condition) %>% summarise(meanCertainty = (sum(Certainty) + .5) / (length(Certainty) + 1))
#  data2 <- merge(data2, temp)
  
#  reliability <- glm(Certainty ~ inverseLogit(meanCertainty), data2, family = binomial())
#  experimentReliability[[i]] <- 1 - reliability$deviance / reliability$null.deviance
#  i <- i + 1
#}

#mean(experimentReliability)

# Predictors of certainty
models <- list(
"Entropy" = glmer(Certainty ~ Entropy + (1 | Condition), data, family = binomial()),
"MAP" = glmer(Certainty ~ MAP + (1 | Condition), data, family = binomial()),
"MAPNoPrior" = glmer(Certainty ~ MAPNoPrior + (1 | Condition), data, family = binomial()),
"TotalCorrect" = glmer(Certainty ~ totalCorrect + (1 | Condition), data, family = binomial()),
"DomainEntropy" = glmer(Certainty ~ DomainEntropy + (1 | Condition), data, family = binomial()),
"LogMAP" = glmer(Certainty ~ log(MAP) + (1 | Condition), data, family = binomial()),
"LogMAPNoPrior" = glmer(Certainty ~ log(MAPNoPrior) + (1 | Condition), data, family = binomial()),
"LogTotalCorrect" = glmer(Certainty ~ log(totalCorrect + 1) + (1 | Condition), data, family = binomial()))

modelComparisons <- data.frame(AIC = sapply(models, AIC),
                               "R^2" = c(cor(meanData$Certainty, meanData$Entropy) ^ 2, cor(meanData$Certainty, meanData$MAP) ^ 2,
                                         cor(meanData$Certainty, meanData$MAPNoPrior) ^ 2, cor(meanData$Certainty, meanData$totalCorrect) ^ 2,
                                         cor(meanData$Certainty, meanData$DomainEntropy) ^ 2, cor(meanData$Certainty, log(meanData$MAP)) ^ 2,
                                         cor(meanData$Certainty, log(meanData$MAPNoPrior)) ^ 2, cor(meanData$Certainty, log(meanData$totalCorrect + 1)) ^ 2,
                                         0),                               
                               "Log Likelihood" = sapply(models, logLik))

modelComparisons <- arrange(modelComparisons, AIC)


# 
#                               Beta = c(summary(a)$coefficients[2], summary(b)$coefficients[2], summary(c)$coefficients[2], summary(d)$coefficients[2],
#                                         summary(e)$coefficients[2], summary(f)$coefficients[2], summary(g)$coefficients[2], summary(h)$coefficients[2], summary(certaintyNull)$coefficients[1]),
#                                "Standard Error" = c(summary(a)$coefficients[2, 2], summary(b)$coefficients[2, 2], summary(c)$coefficients[2, 2], summary(d)$coefficients[2, 2],
#                                                     summary(e)$coefficients[2, 2], summary(f)$coefficients[2, 2], summary(g)$coefficients[2, 2], summary(h)$coefficients[2, 2], summary(certaintyNull)$coefficients[2]),
#                                p = c(summary(a)$coefficients[2, 4], summary(b)$coefficients[2, 4], summary(c)$coefficients[2, 4],
#                                      summary(d)$coefficients[2, 4], summary(e)$coefficients[2, 4], summary(f)$coefficients[2, 4],
#                                      summary(g)$coefficients[2, 4], summary(h)$coefficients[2, 4], summary(certaintyNull)$coefficients[1, 4]))
# 
# 
# print(xtable(modelComparisons, "Predictors of Certainty", "certaintyPredictors", digits = c(0, 0, 1, 2, 1, 2, 2, 3)), include.rownames = FALSE)