library(ggplot2)
library(reshape2)
library(boot)
library(matrixStats)
library(dplyr)
library(tidyr)
library(Hmisc)
library(grid)
library(xtable)
library(lme4)
library(lmerTest)

theme <- theme_bw() + 
  theme(plot.title = element_text(face = "bold", hjust = 0.5),
        axis.title.x = element_text(color = "forestgreen", face = "bold"),
        axis.title.y = element_text(color = "forestgreen", face = "bold"),
        legend.title = element_text(color = "forestgreen", face = "bold"),
        panel.grid.minor.x = element_blank(),
        panel.grid.major.x = element_blank(), aspect.ratio = 1)

theme_set(theme)

###################################################################################################################
# Functions
###################################################################################################################

multiplot <- function(..., plotlist=NULL, file, cols=1, layout=NULL)
{
  library(grid)
  
  # Make a list from the ... arguments and plotlist
  plots <- c(list(...), plotlist)
  
  numPlots = length(plots)
  
  # If layout is NULL, then use 'cols' to determine layout
  if (is.null(layout))
  {
    # Make the panel
    # ncol: Number of columns of plots
    # nrow: Number of rows needed, calculated from # of cols
    layout <- matrix(seq(1, cols * ceiling(numPlots/cols)),
                     ncol = cols, nrow = ceiling(numPlots/cols))
  }
  
  if (numPlots==1) 
  {
    print(plots[[1]])
    
  } 
  else 
  {
    # Set up the page
    grid.newpage()
    pushViewport(viewport(layout = grid.layout(nrow(layout), ncol(layout))))
    
    # Make each plot, in the correct location
    for (i in 1:numPlots)
    {
      # Get the i,j matrix positions of the regions that contain this subplot
      matchidx <- as.data.frame(which(layout == i, arr.ind = TRUE))
      
      print(plots[[i]], vp = viewport(layout.pos.row = matchidx$row,
                                      layout.pos.col = matchidx$col))
    }
  }
}

###################################################################################################################
# Data re-shaping
###################################################################################################################
# Working directory should be set to the same directory as analysis.R
#setwd("Certainty Code and Data\Experiment 4")

bestFitData <- read.csv('bestFitsStudy4.csv')

modelData <- read.csv('modelDataStudy4.csv')
#modelData <- read.csv('modelDataStudy4Simplified.csv')

data <- read.csv("behavioralDataStudy4.csv")

# Create count file
#counts <- ddply(data, .(Condition, Trial), function(df){data.frame(Correct = sum(df$accuracy), Incorrect = sum(df$accuracy == 0), Total = sum(df$accuracy) + sum(df$accuracy == 0))})
#write.csv(counts, "Study4Counts.csv", row.names = TRUE)

# Mean behavioral accuracy for each trial and condition
data <- data %>% group_by(Condition, Trial) %>% mutate(BehavioralAccuracy = mean(accuracy))

# Mean behavioral certainty for each trial and condition
data <- data %>% group_by(Condition, Trial) %>% mutate(BehavioralCertainty = mean(Certainty))

# Local accuracy 1 preceding
data <- data %>% group_by(ID) %>% mutate(localAccuracy1Back = lag(accuracy, default = 0))

# Local accuracy 1 preceding + current
data <- data %>% group_by(ID) %>% mutate(localAccuracy1BackCurrent = lag(accuracy, default = 0) + accuracy)

# Local accuracy 2 preceding
data <- data %>% group_by(ID) %>% mutate(localAccuracy2Back = lag(accuracy, 2, default = 0) + lag(accuracy, default = 0))

# Local accuracy 2 preceding + current
data <- data %>% group_by(ID) %>% mutate(localAccuracy2BackCurrent = lag(accuracy, 2, default = 0) + lag(accuracy, default = 0) + accuracy)

# Local accuracy 3 preceding
data <- data %>% group_by(ID) %>% mutate(localAccuracy3Back = lag(accuracy, 3, default = 0) + lag(accuracy, 2, default = 0) + lag(accuracy, default = 0))

# Local accuracy 3 preceding + current
data <- data %>% group_by(ID) %>% mutate(localAccuracy3BackCurrent = lag(accuracy, 3, default = 0) + lag(accuracy, 2, default = 0) + lag(accuracy, default = 0) + accuracy)

# Local accuracy 4 preceding
data <- data %>% group_by(ID) %>% mutate(localAccuracy4Back = lag(accuracy, 4, default = 0) + lag(accuracy, 3, default = 0) + lag(accuracy, 2, default = 0) + lag(accuracy, default = 0))

# Local accuracy 4 preceding + current
data <- data %>% group_by(ID) %>% mutate(localAccuracy4BackCurrent = lag(accuracy, 4, default = 0) + lag(accuracy, 3, default = 0) + lag(accuracy, 2, default = 0) + lag(accuracy, default = 0) + accuracy)

# Local accuracy 5 preceding
data <- data %>% group_by(ID) %>% mutate(localAccuracy5Back = lag(accuracy, 5, default = 0) + lag(accuracy, 4, default = 0) + lag(accuracy, 3, default = 0) + lag(accuracy, 2, default = 0) + lag(accuracy, default = 0))

# Local accuracy 5 preceding + current
data <- data %>% group_by(ID) %>% mutate(localAccuracy5BackCurrent = lag(accuracy, 5, default = 0) + lag(accuracy, 4, default = 0) + lag(accuracy, 3, default = 0) + lag(accuracy, 2, default = 0) + lag(accuracy, default = 0) + accuracy)

# Total Correct
data <- data %>% group_by(ID) %>% mutate(totalCorrect = cumsum(accuracy))

modelData$PredictedAccuracyBins <- cut(modelData$PredictedAccuracy, breaks = c(0, .1, .2, .3, .4, .5, .6, .7, .8, .9, 1.1), labels = 0:9)
modelData$HypothesisAccuracyBins <- cut(modelData$HypothesisAccuracy, breaks = c(0, .1, .2, .3, .4, .5, .6, .7, .8, .9, 1.1), labels = 0:9)

# Removing RT outliers
data <- filter(data, RT >= 2500)
data <- filter(data, !(abs(RT - median(RT)) > 3 * sd(RT)))

data <- merge(data, modelData, by = c("Condition", "Trial"))

#Inferring participant responses
condition1Answers <-  c(0, 1, 0, 1, 0, 0, 1, 1, 0, 1, 1, 1, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 1, 1)
condition2Answers <-  c(1, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0)
condition3Answers <-  c(1, 1, 1, 0, 1, 1, 0, 1, 1, 1, 0, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, 1)
condition4Answers <-  c(1, 1, 1, 1, 0, 1, 1, 0, 1, 0, 1, 1, 1, 0, 1, 1, 0, 1, 1, 1, 1, 1, 1, 0)
condition5Answers <-  c(1, 0, 1, 1, 0, 1, 0, 0, 1, 1, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0, 1, 1, 0, 1)
condition6Answers <-  c(0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 1, 0, 0, 0, 1, 1, 0, 1)
condition7Answers <-  c(1, 1, 1, 0, 0, 0, 1, 0, 1, 0, 1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 1, 0, 0)
condition8Answers <-  c(1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 0, 1, 0, 1, 1, 1, 1, 0, 0, 1, 0, 0, 1)
condition9Answers <-  c(0, 0, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 0, 0, 1, 1, 1, 0, 0, 1, 0, 0, 0)
condition10Answers <- c(1, 1, 0, 1, 0, 0, 1, 0, 1, 0, 0, 1, 0, 1, 0, 1, 1, 0, 1, 0, 0, 1, 1, 0)

conditions <- c(condition1Answers, condition2Answers, condition3Answers, condition4Answers, condition5Answers,
                condition6Answers, condition7Answers, condition8Answers, condition9Answers, condition10Answers)

data <- mutate(data, Response = ifelse(accuracy, conditions[((as.numeric(Condition) - 1) * 24) + 1], !conditions[((as.numeric(Condition) - 1) * 24) + 1]))

data <- mutate(data, ResponseProbability = ifelse(Response, StimuliPosterior, 1 - StimuliPosterior))

meanData <- data %>% group_by(Trial, Condition) %>% summarise(accuracy = mean(accuracy),
                                                              localAccuracy1Back = mean(localAccuracy1Back),
                                                              localAccuracy2Back = mean(localAccuracy2Back),
                                                              localAccuracy3Back = mean(localAccuracy3Back),
                                                              localAccuracy4Back = mean(localAccuracy4Back),
                                                              localAccuracy5Back = mean(localAccuracy5Back),
                                                              localAccuracy1BackCurrent = mean(localAccuracy1BackCurrent),
                                                              localAccuracy2BackCurrent = mean(localAccuracy2BackCurrent),
                                                              localAccuracy3BackCurrent = mean(localAccuracy3BackCurrent),
                                                              localAccuracy4BackCurrent = mean(localAccuracy4BackCurrent),
                                                              localAccuracy5BackCurrent = mean(localAccuracy5BackCurrent),
                                                              BehavioralCertainty = mean(BehavioralCertainty),
                                                              totalCorrect = mean(totalCorrect),
                                                              Entropy = mean(Entropy),
                                                              MAP = mean(MAP),
                                                              MAPNoPrior = mean(MAPNoPrior),
                                                              ChangeInEntropy = mean(ChangeInEntropy),
                                                              CrossEntropy = mean(CrossEntropy),
                                                              DomainEntropy = mean(DomainEntropy),
                                                              ChangeInDomainEntropy = mean(ChangeInDomainEntropy),
                                                              DomainCrossEntropy = mean(DomainCrossEntropy),
                                                              ResponseProbability = mean(ResponseProbability))

# Reshaping Binary vs. Continuous Data
# data2 <- data %>% group_by(Trial) %>% 
#   summarise(BehavioralCertainty = mean(Certainty, na.rm = TRUE), ContinuousSD = sd(Certainty, na.rm = TRUE), ContinuousN = n()) %>%
#   mutate(ContinuousSE = ContinuousSD / sqrt(ContinuousN), 
#          ContinuousLowCI = BehavioralCertainty - qt(1 - (.05 / 2), ContinuousN - 1) * ContinuousSE,
#          ContinuousHighCI = BehavioralCertainty + qt(1 - (.05 / 2), ContinuousN - 1) * ContinuousSE)
# 
# data3 <- data %>% group_by(Trial) %>% 
#   summarise(BehavioralCertaintyBinary = mean(certainty, na.rm = TRUE), BinarySD = sd(certainty, na.rm = TRUE), BinaryN = n()) %>%
#   mutate(BinarySE = BinarySD / sqrt(BinaryN), 
#          BinaryLowCI = BehavioralCertaintyBinary - qt(1 - (.05 / 2), BinaryN - 1) * BinarySE,
#          BinaryHighCI = BehavioralCertaintyBinary + qt(1 - (.05 / 2), BinaryN - 1) * BinarySE)
# 
# data4 <- merge(data2, data3)
# 
# cor(data4$BehavioralCertainty, data4$BehavioralCertaintyBinary) ^ 2

###################################################################################################################
# Visualizations
###################################################################################################################
# Certainty vs. Local Accuracy 5 Back Current (learners who are self-aware of their certainty) ***C***
v1 <- ggplot(meanData, aes(x = as.numeric(as.character(cut2(localAccuracy5BackCurrent, g = 5, levels.mean = TRUE))), y = BehavioralCertainty)) + 
  stat_summary(fun.y = mean, geom = "point", shape = 5, size = 5, color = "blue") +
  stat_summary(fun.data = mean_cl_boot, geom = "pointrange", color = "blue") +
  geom_point(aes(x = localAccuracy5BackCurrent, y = BehavioralCertainty), alpha = .1) +
  ggtitle(expression(R^{2} ~ " = .77")) +
  labs(x = "Local Accuracy 5 Back Current", y = "Certainty") +
  expand_limits(y = 0, y = 1) +
  scale_y_continuous(expand = c(0, 0))

# Certainty vs. Log Total Correct
v2 <- ggplot(meanData, aes(x = as.numeric(as.character(cut2(log(totalCorrect + 1), g = 5, levels.mean = TRUE))), y = BehavioralCertainty)) + 
  stat_summary(fun.y = mean, geom = "point", shape = 5, size = 5, color = "blue") +
  stat_summary(fun.data = mean_cl_boot, geom = "pointrange", color = "blue") +
  geom_point(aes(x = log(totalCorrect + 1), y = BehavioralCertainty), alpha = .1) +
  ggtitle(expression(R^{2} ~ " = .56")) +
  labs(x = "Log Total Correct", y = "Certainty") +
  expand_limits(y = 0, y = 1) +
  scale_y_continuous(expand = c(0, 0))

# Certainty vs. Domain Entropy
v3 <- ggplot(meanData, aes(x = as.numeric(as.character(cut2(DomainEntropy, g = 5, levels.mean = TRUE))), y = BehavioralCertainty)) +
  stat_summary(fun.y = mean, geom = "point", shape = 5, size = 5, color = "blue") +
  stat_summary(fun.data = mean_cl_boot, geom = "pointrange", color = "blue") +
  geom_point(aes(x = DomainEntropy, y = BehavioralCertainty), alpha = .1) +
  ggtitle(expression(R^{2} ~ " = .69")) +
  labs(x = "Domain Entropy", y = "Certainty") +
  expand_limits(y = 0, y = 1) +
  scale_y_continuous(expand = c(0, 0))

# Certainty by Log Trial (learner who waits exhaustively for data) ***B***
v4 <- ggplot(meanData, aes(x = as.numeric(as.character(cut2(log(Trial), g = 5, levels.mean = TRUE))), y = BehavioralCertainty)) + 
  stat_summary(fun.y = mean, geom = "point", shape = 5, size = 5, color = "blue") +
  stat_summary(fun.data = mean_cl_boot, geom = "pointrange", color = "blue") +
  geom_point(aes(x = log(Trial), y = BehavioralCertainty), alpha = .1) +
  ggtitle(expression(R^{2} ~ " = .44")) +
  labs(x = "Log Trial", y = "Certainty") +
  expand_limits(y = 0, y = 1) +
  scale_y_continuous(expand = c(0, 0))

# Certainty By Log Max Likelihood (learners without a simplicity bias) ***E***
v5 <- ggplot(meanData, aes(x = as.numeric(as.character(cut2(log(MAPNoPrior), g = 5, levels.mean = TRUE))), y = BehavioralCertainty)) + 
  stat_summary(fun.y = mean, geom = "point", shape = 5, size = 5, color = "blue") +
  stat_summary(fun.data = mean_cl_boot, geom = "pointrange", color = "blue") +
  geom_point(aes(x = log(MAPNoPrior), y = BehavioralCertainty), alpha = .1) +
  ggtitle(expression(R^{2} ~ " = .41")) +
  labs(x = "Log Max Likelihood", y = "Certainty") +
  expand_limits(y = 0, y = 1) +
  scale_y_continuous(expand = c(0, 0))

# Certainty vs. Entropy (Ideal Learner) ***A***
v6 <- ggplot(meanData, aes(x = as.numeric(as.character(cut2(Entropy, g = 5, levels.mean = TRUE))), y = BehavioralCertainty)) +
  stat_summary(fun.y = mean, geom = "point", shape = 5, size = 5, color = "blue") +
  stat_summary(fun.data = mean_cl_boot, geom = "pointrange", color = "blue") +
  geom_point(aes(x = Entropy, y = BehavioralCertainty), alpha = .1) +
  ggtitle(expression(R^{2} ~ " = .52")) +
  labs(x = "Entropy", y = "Certainty") +
  expand_limits(y = 0, y = 1) +
  scale_y_continuous(expand = c(0, 0))

# Certainty By MAP (learners who only have a single hypothesis in mind) ***D***
v7 <- ggplot(meanData, aes(x = as.numeric(as.character(cut2(MAP, g = 5, levels.mean = TRUE))), y = BehavioralCertainty)) + 
  stat_summary(fun.y = mean, geom = "point", shape = 5, size = 5, color = "blue") +
  stat_summary(fun.data = mean_cl_boot, geom = "pointrange", color = "blue") +
  geom_point(aes(x = MAP, y = BehavioralCertainty), alpha = .1) +
  ggtitle(expression(R^{2} ~ " = .33")) +
  labs(x = "MAP Hypothesis", y = "Certainty") +
  expand_limits(y = 0, y = 1) +
  scale_y_continuous(expand = c(0, 0))

# Certainty vs. Change in Entropy
v8 <- ggplot(meanData, aes(x = as.numeric(as.character(cut2(ChangeInEntropy, g = 5, levels.mean = TRUE))), y = BehavioralCertainty)) +
  stat_summary(fun.y = mean, geom = "point", shape = 5, size = 5, color = "blue") +
  stat_summary(fun.data = mean_cl_boot, geom = "pointrange", color = "blue") +
  geom_point(aes(x = ChangeInEntropy, y = BehavioralCertainty), alpha = .1) +
  ggtitle(expression(R^{2} ~ " = .16")) +
  labs(x = "Change in Entropy", y = "Certainty") +
  expand_limits(y = 0, y = 1) +
  scale_y_continuous(expand = c(0, 0))

# Certainty vs. Change in Domain Entropy
v9 <- ggplot(meanData, aes(x = as.numeric(as.character(cut2(ChangeInDomainEntropy, g = 5, levels.mean = TRUE))), y = BehavioralCertainty)) +
  stat_summary(fun.y = mean, geom = "point", shape = 5, size = 5, color = "blue") +
  stat_summary(fun.data = mean_cl_boot, geom = "pointrange", color = "blue") +
  geom_point(aes(x = ChangeInDomainEntropy, y = BehavioralCertainty), alpha = .1) +
  ggtitle(expression(R^{2} ~ " = .16")) +
  labs(x = "Change in Domain", y = "Certainty") +
  expand_limits(y = 0, y = 1) +
  scale_y_continuous(expand = c(0, 0))

# Certainty vs. Cross Entropy
v10 <- ggplot(meanData, aes(x = as.numeric(as.character(cut2(CrossEntropy, g = 5, levels.mean = TRUE))), y = BehavioralCertainty)) +
  stat_summary(fun.y = mean, geom = "point", shape = 5, size = 5, color = "blue") +
  stat_summary(fun.data = mean_cl_boot, geom = "pointrange", color = "blue") +
  geom_point(aes(x = CrossEntropy, y = BehavioralCertainty), alpha = .1) +
  ggtitle(expression(R^{2} ~ " = .21")) +
  labs(x = "Cross Entropy", y = "Certainty") +
  expand_limits(y = 0, y = 1) +
  scale_y_continuous(expand = c(0, 0))

# Certainty vs. Log Response Probability
v11 <- ggplot(meanData, aes(x = as.numeric(as.character(cut2(log(ResponseProbability), g = 5, levels.mean = TRUE))), y = BehavioralCertainty)) +
  stat_summary(fun.y = mean, geom = "point", shape = 5, size = 5, color = "blue") +
  stat_summary(fun.data = mean_cl_boot, geom = "pointrange", color = "blue") +
  geom_point(aes(x = log(ResponseProbability), y = BehavioralCertainty), alpha = .1) +
  ggtitle(expression(R^{2} ~ " = .14")) +
  labs(x = "Log Response Probability", y = "Certainty") +
  expand_limits(y = 0, y = 1) +
  scale_y_continuous(expand = c(0, 0))

# Certainty vs. Domain Cross Entropy
v12 <- ggplot(meanData, aes(x = as.numeric(as.character(cut2(DomainCrossEntropy, g = 5, levels.mean = TRUE))), y = BehavioralCertainty)) +
  stat_summary(fun.y = mean, geom = "point", shape = 5, size = 5, color = "blue") +
  stat_summary(fun.data = mean_cl_boot, geom = "pointrange", color = "blue") +
  geom_point(aes(x = DomainCrossEntropy, y = BehavioralCertainty), alpha = .1) +
  ggtitle(expression(R^{2} ~ " = .03")) +
  labs(x = "Domain Cross Entropy", y = "Certainty") +
  expand_limits(y = 0, y = 1) +
  scale_y_continuous(expand = c(0, 0))

multiplot(v1, v4, v7, v10, 
          v2, v5, v8, v11, 
          v3, v6, v9, v12, 
          cols = 3)

# Continuous Certainty (Exp 4) by Binary Certainty (Exp 1)
ggplot(data4, aes(x = BehavioralCertaintyBinary, y = BehavioralCertainty)) + 
  geom_point() + 
  geom_errorbar(aes(ymin = ContinuousLowCI, 
                    ymax = ContinuousHighCI), width = 0.05) +
  geom_errorbarh(aes(xmin = BinaryLowCI, 
                     xmax = BinaryHighCI), height = 0.2) +
  ggtitle("Continuous Certainty by Binary Certainty") +
  geom_abline(slope = 4, intercept = 1, linetype = "dotted") +
  xlim(0, 1) + 
  ylim(1, 5)

# Certainty and Accuracy
ggplot(data, aes(x = accuracy, y = Certainty)) + 
  stat_summary(geom = 'point') +
  ggtitle("Certainty and Accuracy")

# Behavioral Accuracy Binned by Model Accuracy
ggplot(data, aes(x = as.numeric(as.character(cut2(PredictedAccuracy, g = 5, levels.mean = TRUE))), y = BehavioralAccuracy)) + 
  stat_summary(fun.y = mean, geom = "point", shape = 5, size = 5, color = "blue") +
  stat_summary(fun.data = mean_cl_boot, geom = "pointrange", color = "blue") +
  geom_point(aes(x = PredictedAccuracy, y = BehavioralAccuracy), alpha = .005) +
  expand_limits(x = 0, y = 0) +
  expand_limits(x = 1, y = 1) +
  scale_x_continuous(expand = c(0, 0)) + 
  scale_y_continuous(expand = c(0, 0)) +
  geom_abline(slope = 1, linetype = "dotted") +
  theme(axis.title = element_text(size = 14, face = "bold")) +
  labs(x = "Model Accuracy", y = "Behavioral Accuracy") +
  ggtitle("Model Predicts Accuracy")

# Certainty and Accuracy by Trial
ggplot(data, aes(x = Trial, y = Certainty / 5)) + 
  stat_summary(fun.y = mean, geom = "point", shape = 5, size = 4, fill = "green") + 
  stat_summary(fun.data = mean_cl_boot, geom = "pointrange", color = "green") +
  stat_summary(aes(x = Trial, y = accuracy), fun.y = mean, geom = "point", shape = 1, size = 4, color = "red") + 
  stat_summary(aes(x = Trial, y = accuracy), fun.data = mean_cl_boot, geom = "errorbar", color = "red") +
  ggtitle("Certainty and Accuracy by Trial") +
  ylim(0, 1)

# Certainty and Accuracy by Condition
ggplot(data, aes(x = Trial, y = Certainty / 5)) + 
  stat_summary(fun.y = mean, geom = "point", shape = 1, color = "blue") + 
  stat_summary(fun.data = mean_cl_boot, geom = "linerange") +
  stat_summary(aes(x = Trial, y = accuracy), fun.y = mean, geom = "point", color = "red") + 
  stat_summary(aes(x = Trial, y = accuracy), fun.data = mean_cl_boot, geom = "linerange") +
  ggtitle("Certainty and Accuracy by Condition") +
  labs(color = "Condition", y = "") +
  ylim(0, 1) +
  scale_x_continuous(breaks = c(1, 8, 16, 24)) +
  facet_wrap(~Condition)

# Certainty and Accuracy by Subject
ggplot(data, aes(x = Trial, y = Certainty)) + 
  stat_summary(fun.y = mean, geom = "point", shape = 5, size = 5, fill = "blue") + 
  stat_summary(fun.data = mean_cl_boot, geom = "pointrange", color = "blue") +
  stat_summary(aes(x = Trial, y = accuracy), fun.y = mean, geom = "point", shape = 7, size = 5, color = "red") + 
  stat_summary(aes(x = Trial, y = accuracy), fun.data = mean_cl_boot, geom = "pointrange", color = "red") +
  ggtitle("Certainty and Accuracy by Condition") +
  labs(color = "Condition", y = "") +
  scale_color_manual(values = c("Women" = '#ff00ff','Men' = '#3399ff')) +
  ylim(0, 1) +
  facet_wrap(~ID)

# Behavioral Accuracy By Model Accuracy
ggplot(data, aes(x = PredictedAccuracy, y = BehavioralAccuracy)) + 
  stat_summary(fun.y = mean, geom = "point", shape = 1, size = 4, color = "blue") +
  ggtitle("Behavioral Accuracy By Model Accuracy") +
  geom_abline(slope = 1, intercept = 0) +
  xlim(0, 1) +
  facet_wrap(~Condition)

# Behavioral Accuracy Binned by Model Accuracy
ggplot(data, aes(x = HypothesisAccuracyBins, y = BehavioralAccuracy)) + 
  geom_boxplot() + 
  stat_summary(fun.y = mean, geom = "point", shape = 5, size = 4) + 
  stat_summary(fun.data = mean_cl_boot, geom = "pointrange") +
  ylim(0, 1) +
  ggtitle("Behavioral Accuracy Binned by Model Accuracy") +
  geom_point() + 
  geom_jitter(position = position_jitter(w = .4, h = .1)) +
  facet_wrap(~Condition)

# Behavioral Certainty Binned by Model Accuracy
ggplot(data, aes(x = ModelAccuracyBins, y = BehavioralCertainty)) + 
  geom_boxplot() + 
  geom_point() + 
  geom_jitter(position = position_jitter(w = .4, h = .1)) +
  stat_summary(fun.y = mean, geom = "point", shape = 5, size = 4) + 
  stat_summary(fun.data = mean_cl_boot, geom = "pointrange") +
  ylim(0, 1) +
  ggtitle("Behavioral Certainty Binned by Model Accuracy")

# RT by Trial
ggplot(data, aes(x = Trial, y = RT, color = Certainty)) +
  stat_summary(fun.y = mean, geom = "point") +
  geom_smooth(method = "lm", alpha = 0.1) + #, formula = y ~ x + Condition) + 
  ggtitle("RT by Trial") +
  facet_wrap(~Condition)

# RT by Condition
ggplot(data, aes(x = Condition, y = RT, fill = Certainty)) + 
  stat_summary(fun.y = mean, geom = "bar", position = position_dodge(.9)) +
  stat_summary(fun.data = mean_cl_boot, geom = "pointrange", position = position_dodge(.9)) +
  ggtitle("RT by Condition")

# RT/Certainty
ggplot(data, aes(x = Certainty, y = RT)) + 
  stat_summary(fun.y = mean, geom = "bar") +
  stat_summary(fun.data = mean_cl_boot, geom = "pointrange") +
  ggtitle("RT by Certainty")

# RT/Accuracy
ggplot(data, aes(x = Accuracy, y = RT, fill = Certainty)) + 
  stat_summary(fun.y = mean, geom = "bar", position = position_dodge(.9)) +
  stat_summary(fun.data = mean_cl_boot, geom = "pointrange", position = position_dodge(.9)) +
  ggtitle("RT by Accuracy")

###################################################################################################################
# Regressions
###################################################################################################################
accuracyNull <- lmer(accuracy ~ (1 | ID) + (1 | Condition), data)
certaintyNull <- lmer(Certainty ~ (1 | ID) + (1 | Condition), data)

modelPredictsAccuracy <- lmer(accuracy ~ scale(PredictedAccuracy) + (1 | ID) + (1 | Condition), data)
summary(modelPredictsAccuracy)

summary(lmer(RT ~ Certainty + (1 | ID),data))

summary(lmer(RT ~ accuracy + (1 | ID),data))

summary(lmer(RT ~ NumberAccurate + (1 | ID),data))

summary(lmer(RT ~ Certainty * accuracy * Trial + (1 | ID),data))

# Subject reliability
#i <- 1
#experimentReliability <- 0

#while(i <= 1000) 
#{
#  data1 <- data %>% sample_n(nrow(data) / 2, replace = FALSE)
#  data2 <- anti_join(data, data1)
#  temp <- data1 %>% group_by(Trial, Condition) %>% summarise(meanCertainty = (sum(Certainty) + .5) / (length(Certainty) + 1))
#  data2 <- merge(data2, temp)
  
#  reliability <- lm(Certainty ~ meanCertainty, data2)
  
#  experimentReliability[[i]] <- summary(reliability)$adj.r.squared
#  i <- i + 1
#}

#mean(experimentReliability)

models <- list(
  "Entropy" = lmer(Certainty ~ scale(Entropy) + (1 | ID) + (1 | Condition), data),
  "Trial" = lmer(Certainty ~ scale(Trial) + (1 | ID) + (1 | Condition), data),
  "Accuracy" = lmer(Certainty ~ scale(accuracy) + (1 | ID) + (1 | Condition), data),
  "LocalAccuracy1BackCurrent" = lmer(Certainty ~ scale(localAccuracy1BackCurrent) + (1 | ID) + (1 | Condition), data),
  "LocalAccuracy1Back" = lmer(Certainty ~ scale(localAccuracy1Back) + (1 | ID) + (1 | Condition), data),
  "LocalAccuracy2BackCurrent" = lmer(Certainty ~ scale(localAccuracy2BackCurrent) + (1 | ID) + (1 | Condition), data),
  "LocalAccuracy2Back" = lmer(Certainty ~ scale(localAccuracy2Back) + (1 | ID) + (1 | Condition), data),
  "LocalAccuracy3BackCurrent" = lmer(Certainty ~ scale(localAccuracy3BackCurrent) + (1 | ID) + (1 | Condition), data),
  "LocalAccuracy3Back" = lmer(Certainty ~ scale(localAccuracy3Back) + (1 | ID) + (1 | Condition), data),
  "LocalAccuracy4BackCurrent" = lmer(Certainty ~ scale(localAccuracy4BackCurrent) + (1 | ID) + (1 | Condition), data),
  "LocalAccuracy4Back" = lmer(Certainty ~ scale(localAccuracy4Back) + (1 | ID) + (1 | Condition), data),
  "LocalAccuracy5BackCurrent" = lmer(Certainty ~ scale(localAccuracy5BackCurrent) + (1 | ID) + (1 | Condition), data),
  "LocalAccuracy5Back" = lmer(Certainty ~ scale(localAccuracy5Back) + (1 | ID) + (1 | Condition), data),
  "MAP" = lmer(Certainty ~ scale(MAP) + (1 | ID) + (1 | Condition), data),
  "MAPNoPrior" = lmer(Certainty ~ scale(MAPNoPrior) + (1 | ID) + (1 | Condition), data),
  "TotalCorrect" = lmer(Certainty ~ scale(totalCorrect) + (1 | ID) + (1 | Condition), data),
  "ChangeInEntropy" = lmer(Certainty ~ scale(ChangeInEntropy) + (1 | ID) + (1 | Condition), data),
  "CrossEntropy" = lmer(Certainty ~ scale(CrossEntropy) + (1 | ID) + (1 | Condition), data),
  "DomainEntropy" = lmer(Certainty ~ scale(DomainEntropy) + (1 | ID) + (1 | Condition), data),
  "ChangeInDomainEntropy" = lmer(Certainty ~ scale(ChangeInDomainEntropy)  + (1 | ID) + (1 | Condition), data),
  "DomainCrossEntropy" = lmer(Certainty ~ scale(DomainCrossEntropy) + (1 | ID) + (1 | Condition), data),
  "Response Probability" = lmer(Certainty ~ scale(ResponseProbability) + (1 | ID) + (1 | Condition), data),
  "LogTrial" = lmer(Certainty ~ scale(log(Trial)) + (1 | ID) + (1 | Condition), data),
  "LogAccuracy" = lmer(Certainty ~ scale(log(accuracy + 1)) + (1 | ID) + (1 | Condition), data),
  "LogLocalAccuracy1BackCurrent" = lmer(Certainty ~ scale(log(localAccuracy1BackCurrent + 1)) + (1 | ID) + (1 | Condition), data),
  "LogLocalAccuracy1Back" = lmer(Certainty ~ scale(log(localAccuracy1Back + 1)) + (1 | ID) + (1 | Condition), data),
  "LogLocalAccuracy2BackCurent" = lmer(Certainty ~ scale(log(localAccuracy2BackCurrent + 1)) + (1 | ID) + (1 | Condition), data),
  "LogLocalAccuracy2Back" = lmer(Certainty ~ scale(log(localAccuracy2Back + 1)) + (1 | ID) + (1 | Condition), data),
  "LogLocalAccuracy3BackCurrent" = lmer(Certainty ~ scale(log(localAccuracy3BackCurrent + 1)) + (1 | ID) + (1 | Condition), data),
  "LogLocalAccuracy3Back" = lmer(Certainty ~ scale(log(localAccuracy3Back + 1)) + (1 | ID) + (1 | Condition), data),
  "LogLocalAccuracy4BackCurrent" = lmer(Certainty ~ scale(log(localAccuracy4BackCurrent + 1)) + (1 | ID) + (1 | Condition), data),
  "LogLocalAccuracy4Back" = lmer(Certainty ~ scale(log(localAccuracy4Back + 1)) + (1 | ID) + (1 | Condition), data),
  "LogLocalAccuracy5BackCurrent" = lmer(Certainty ~ scale(log(localAccuracy5BackCurrent + 1)) + (1 | ID) + (1 | Condition), data),
  "LogLocalAccuracy5Back" = lmer(Certainty ~ scale(log(localAccuracy5Back + 1)) + (1 | ID) + (1 | Condition), data),
  "LogMAP" = lmer(Certainty ~ scale(log(MAP)) + (1 | ID) + (1 | Condition), data),
  "LogMAPNoPrior" = lmer(Certainty ~ scale(log(MAPNoPrior)) + (1 | ID) + (1 | Condition), data),
  "LogTotalCorrect" = lmer(Certainty ~ scale(log(totalCorrect + 1)) + (1 | ID) + (1 | Condition), data),
  "LogResponseProbability" = lmer(Certainty ~ scale(log(ResponseProbability)) + (1 | ID) + (1 | Condition), data),
  "CertaintyNull" = certaintyNull)

modelComparisons <- data.frame(AIC = sapply(models, AIC),
                               "R^2" = c(cor(meanData$BehavioralCertainty, meanData$Entropy) ^ 2, cor(meanData$BehavioralCertainty, meanData$Trial) ^ 2, cor(meanData$BehavioralCertainty, meanData$accuracy) ^ 2, 
                                         cor(meanData$BehavioralCertainty, meanData$localAccuracy1BackCurrent) ^ 2, cor(meanData$BehavioralCertainty, meanData$localAccuracy1Back) ^ 2, 
                                         cor(meanData$BehavioralCertainty, meanData$localAccuracy2BackCurrent) ^ 2, cor(meanData$BehavioralCertainty, meanData$localAccuracy2Back) ^ 2, 
                                         cor(meanData$BehavioralCertainty, meanData$localAccuracy3BackCurrent) ^ 2, cor(meanData$BehavioralCertainty, meanData$localAccuracy3Back) ^ 2, 
                                         cor(meanData$BehavioralCertainty, meanData$localAccuracy4BackCurrent) ^ 2, cor(meanData$BehavioralCertainty, meanData$localAccuracy4Back) ^ 2, 
                                         cor(meanData$BehavioralCertainty, meanData$localAccuracy5BackCurrent) ^ 2, cor(meanData$BehavioralCertainty, meanData$localAccuracy5Back) ^ 2, 
                                         cor(meanData$BehavioralCertainty, meanData$MAP) ^ 2, cor(meanData$BehavioralCertainty, meanData$MAPNoPrior) ^ 2, cor(meanData$BehavioralCertainty, meanData$totalCorrect)  ^ 2,
                                         cor(meanData$BehavioralCertainty, meanData$ChangeInEntropy) ^ 2, cor(meanData$BehavioralCertainty, meanData$CrossEntropy) ^ 2, cor(meanData$BehavioralCertainty, meanData$DomainEntropy) ^ 2,
                                         cor(meanData$BehavioralCertainty, meanData$ChangeInDomainEntropy) ^ 2, cor(meanData$BehavioralCertainty, meanData$DomainCrossEntropy) ^ 2, cor(meanData$BehavioralCertainty, meanData$ResponseProbability) ^ 2,
                                         cor(meanData$BehavioralCertainty, log(meanData$Trial)) ^ 2, cor(meanData$BehavioralCertainty, log(meanData$accuracy + 1)) ^ 2, 
                                         cor(meanData$BehavioralCertainty, log(meanData$localAccuracy1BackCurrent + 1)) ^ 2, cor(meanData$BehavioralCertainty, log(meanData$localAccuracy1Back + 1)) ^ 2, 
                                         cor(meanData$BehavioralCertainty, log(meanData$localAccuracy2BackCurrent + 1)) ^ 2, cor(meanData$BehavioralCertainty, log(meanData$localAccuracy2Back + 1)) ^ 2, 
                                         cor(meanData$BehavioralCertainty, log(meanData$localAccuracy3BackCurrent + 1)) ^ 2, cor(meanData$BehavioralCertainty, log(meanData$localAccuracy3Back + 1)) ^ 2, 
                                         cor(meanData$BehavioralCertainty, log(meanData$localAccuracy4BackCurrent + 1)) ^ 2, cor(meanData$BehavioralCertainty, log(meanData$localAccuracy4Back + 1)) ^ 2, 
                                         cor(meanData$BehavioralCertainty, log(meanData$localAccuracy5BackCurrent + 1)) ^ 2, cor(meanData$BehavioralCertainty, log(meanData$localAccuracy5Back + 1)) ^ 2,
                                         cor(meanData$BehavioralCertainty, log(meanData$MAP)) ^ 2, cor(meanData$BehavioralCertainty, log(meanData$MAPNoPrior)) ^ 2, cor(meanData$BehavioralCertainty, log(meanData$totalCorrect + 1)) ^ 2,
                                         cor(meanData$BehavioralCertainty, log(meanData$ResponseProbability))  ^ 2, 0),
                               "Log Likelihood" = sapply(models, logLik))

modelComparisons <- arrange(modelComparisons, AIC)


#                                Beta = c(summary(a)$coefficients[2], summary(b)$coefficients[2], summary(c)$coefficients[2], summary(d)$coefficients[2],
#                                         summary(e)$coefficients[2], summary(f)$coefficients[2], summary(g)$coefficients[2], summary(h)$coefficients[2],
#                                         summary(i)$coefficients[2], summary(j)$coefficients[2], summary(k)$coefficients[2], summary(l)$coefficients[2],
#                                         summary(m)$coefficients[2], summary(n)$coefficients[2], summary(o)$coefficients[2], summary(p)$coefficients[2],
#                                         summary(q)$coefficients[2], summary(r)$coefficients[2], summary(s)$coefficients[2], summary(t)$coefficients[2],
#                                         summary(u)$coefficients[2], summary(v)$coefficients[2], summary(w)$coefficients[2], summary(x)$coefficients[2],
#                                         summary(y)$coefficients[2], summary(z)$coefficients[2], summary(aa)$coefficients[2], summary(bb)$coefficients[2],
#                                         summary(cc)$coefficients[2], summary(dd)$coefficients[2], summary(ee)$coefficients[2], summary(ff)$coefficients[2],
#                                         summary(gg)$coefficients[2], summary(hh)$coefficients[2], summary(ii)$coefficients[2], summary(jj)$coefficients[2], 
#                                         summary(kk)$coefficients[2], summary(ll)$coefficients[2], summary(certaintyNull)$coefficients[1]),
#                               "Standard Error" = c(summary(a)$coefficients[2, 2], summary(b)$coefficients[2, 2], summary(c)$coefficients[2, 2], summary(d)$coefficients[2, 2],
#                                                    summary(e)$coefficients[2, 2], summary(f)$coefficients[2, 2], summary(g)$coefficients[2, 2], summary(h)$coefficients[2, 2],
#                                                    summary(i)$coefficients[2, 2], summary(j)$coefficients[2, 2], summary(k)$coefficients[2, 2], summary(l)$coefficients[2, 2],
#                                                    summary(m)$coefficients[2, 2], summary(n)$coefficients[2, 2], summary(o)$coefficients[2, 2], summary(p)$coefficients[2, 2],
#                                                    summary(q)$coefficients[2, 2], summary(r)$coefficients[2, 2], summary(s)$coefficients[2, 2], summary(t)$coefficients[2, 2],
#                                                    summary(u)$coefficients[2, 2], summary(v)$coefficients[2, 2], summary(w)$coefficients[2, 2], summary(x)$coefficients[2, 2],
#                                                    summary(y)$coefficients[2, 2], summary(z)$coefficients[2, 2], summary(aa)$coefficients[2, 2], summary(bb)$coefficients[2, 2],
#                                                    summary(cc)$coefficients[2, 2], summary(dd)$coefficients[2, 2], summary(ee)$coefficients[2, 2], summary(ff)$coefficients[2, 2],
#                                                    summary(gg)$coefficients[2, 2], summary(hh)$coefficients[2, 2], summary(ii)$coefficients[2, 2], summary(jj)$coefficients[2, 2], 
#                                                    summary(kk)$coefficients[2, 2], summary(ll)$coefficients[2, 2], summary(certaintyNull)$coefficients[2]),
#                                p = c(summary(a)$coefficients[2, 5], summary(b)$coefficients[2, 5], summary(c)$coefficients[2, 5],
#                                      summary(d)$coefficients[2, 5], summary(e)$coefficients[2, 5], summary(f)$coefficients[2, 5],
#                                      summary(g)$coefficients[2, 5], summary(h)$coefficients[2, 5], summary(i)$coefficients[2, 5],
#                                      summary(j)$coefficients[2, 5], summary(k)$coefficients[2, 5], summary(l)$coefficients[2, 5],
#                                      summary(m)$coefficients[2, 5], summary(n)$coefficients[2, 5], summary(o)$coefficients[2, 5],
#                                      summary(p)$coefficients[2, 5], summary(q)$coefficients[2, 5], summary(r)$coefficients[2, 5],
#                                      summary(s)$coefficients[2, 5], summary(t)$coefficients[2, 5], summary(u)$coefficients[2, 5],
#                                      summary(v)$coefficients[2, 5], summary(w)$coefficients[2, 5], summary(x)$coefficients[2, 5],
#                                      summary(y)$coefficients[2, 5], summary(z)$coefficients[2, 5], summary(aa)$coefficients[2, 5],
#                                      summary(bb)$coefficients[2, 5], summary(cc)$coefficients[2, 5], summary(dd)$coefficients[2, 5],
#                                      summary(ee)$coefficients[2, 5], summary(ff)$coefficients[2, 5], summary(gg)$coefficients[2, 5], 
#                                      summary(hh)$coefficients[2, 5], summary(ii)$coefficients[2, 5], summary(jj)$coefficients[2, 5], 
#                                      summary(kk)$coefficients[2, 5], summary(ll)$coefficients[2, 5], 0))
# 
# print(xtable(modelComparisons, "Predictors of Certainty", "certaintyPredictors", digits = c(0, 0, 1, 2, 1, 2, 2, 3)), include.rownames = FALSE)